BytellHashTable

Struct BytellHashTable 

pub struct BytellHashTable<TKey, TValue, TConfig>
where TKey: Eq + Hash, TConfig: BytellConfig,
{ /* private fields */ }
The bytell hash table.

Implementations

Source

impl<TKey, TValue, TConfig> BytellHashTable<TKey, TValue, TConfig>
where TKey: Eq + Hash, TConfig: BytellConfig,

pub fn new() -> Self
where TConfig: Default,

Creates a new BytellHashTable with the default configuration.

pub const fn with_config(config: TConfig) -> Self

Creates a new BytellHashTable with the specified configuration.

pub fn with_capacity(capacity: Length) -> Result<Self, HashTableError>
where TConfig: Default,

Creates a new BytellHashTable with the specified capacity and the default configuration.

Errors

For details see HashTableError.

pub fn with_capacity_and_config( capacity: Length, config: TConfig, ) -> Result<Self, HashTableError>

Creates a new BytellHashTable expected to support passed number_of_items.

Notes

This method will almost surely overallocate, since it takes config.load_factor() into account.

Errors

For details see HashTableError.

pub fn shrink_to_fit(&mut self) -> Result<(), HashTableError>

Reduces the memory of the current table, if possible.

Notes

If the number of underlying blocks cannot be reduced, it does nothing.

Otherwise it tries to reduce the number of blocks to the minimal possible, creates a new table with that layout, and moves (rehases) items into it. Then overwrites self with the new table.

Errors

For details see HashTableError.

pub const fn length(&self) -> Length

Returns the length of the BytellHashTable.

pub const fn capacity(&self) -> PowerOfTwo32

Returns the capacity of the BytellHashTable.