pub struct BytellHashTable<TKey, TValue, TConfig>{ /* private fields */ }Expand description
The bytell hash table.
Implementations
Sourceimpl<TKey, TValue, TConfig> BytellHashTable<TKey, TValue, TConfig>
impl<TKey, TValue, TConfig> BytellHashTable<TKey, TValue, TConfig>
Sourcepub fn new() -> Selfwhere
TConfig: Default,
pub fn new() -> Selfwhere
TConfig: Default,
Creates a new BytellHashTable with the default configuration.
Sourcepub const fn with_config(config: TConfig) -> Self
pub const fn with_config(config: TConfig) -> Self
Creates a new BytellHashTable with the specified configuration.
Sourcepub fn with_capacity(capacity: Length) -> Result<Self, HashTableError>where
TConfig: Default,
pub fn with_capacity(capacity: Length) -> Result<Self, HashTableError>where
TConfig: Default,
Creates a new BytellHashTable with the specified capacity and the default configuration.
Errors
For details see HashTableError.
Sourcepub fn with_capacity_and_config(
capacity: Length,
config: TConfig,
) -> Result<Self, HashTableError>
pub fn with_capacity_and_config( capacity: Length, config: TConfig, ) -> Result<Self, HashTableError>
Creates a new BytellHashTable expected to support passed number_of_items.
Notes
This method will almost surely overallocate, since it takes config.load_factor()
into account.
Errors
For details see HashTableError.
Sourcepub fn shrink_to_fit(&mut self) -> Result<(), HashTableError>
pub fn shrink_to_fit(&mut self) -> Result<(), HashTableError>
Reduces the memory of the current table, if possible.
Notes
If the number of underlying blocks cannot be reduced, it does nothing.
Otherwise it tries to reduce the number of blocks to the
minimal possible, creates a new table with that layout,
and moves (rehases) items into it. Then overwrites self
with the new table.
Errors
For details see HashTableError.
Sourcepub const fn length(&self) -> Length
pub const fn length(&self) -> Length
Returns the length of the BytellHashTable.
Sourcepub const fn capacity(&self) -> PowerOfTwo32
pub const fn capacity(&self) -> PowerOfTwo32
Returns the capacity of the BytellHashTable.